@@ -1,18 +1,18 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
+static const MVMREPROps Semaphore_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st  = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st  = MVM_gc_allocate_stable(tc, &Semaphore_this_repr, HOW);
 
-    MVMROOT(tc, st, {
+    MVMROOT(tc, st) {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
         st->size = sizeof(MVMSemaphore);
-    });
+    }
 
     return st->WHAT;
 }
@@ -26,15 +26,21 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 static void set_int(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 value) {
     MVMSemaphoreBody *body = (MVMSemaphoreBody *)data;
     int r;
-    if ((r = uv_sem_init(&body->sem, (MVMuint32) value)) < 0)
+    body->sem = MVM_malloc(sizeof(uv_sem_t));
+    if ((r = uv_sem_init(body->sem, (MVMuint32) value)) < 0) {
+        MVM_free_null(body->sem);
         MVM_exception_throw_adhoc(tc, "Failed to initialize Semaphore: %s",
             uv_strerror(r));
+    }
 }
 
 /* Called by the VM in order to free memory associated with this object. */
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMSemaphore *sem = (MVMSemaphore *)obj;
-    uv_sem_destroy(&sem->body.sem);
+    if (sem->body.sem) {
+        uv_sem_destroy(sem->body.sem);
+        MVM_free(sem->body.sem);
+    }
 }
 
 static const MVMStorageSpec storage_spec = {
@@ -64,10 +70,10 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
 
 /* Initializes the representation. */
 const MVMREPROps * MVMSemaphore_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &Semaphore_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps Semaphore_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -103,20 +109,30 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "Semaphore", /* name */
     MVM_REPR_ID_Semaphore,
-    0, /* Semaphore */
     NULL, /* unmanaged_size */
     NULL, /* describe_refs */
 };
 
 MVMint64 MVM_semaphore_tryacquire(MVMThreadContext *tc, MVMSemaphore *sem) {
-    int r = uv_sem_trywait(&sem->body.sem);
+    int r;
+    unsigned int interval_id = MVM_telemetry_interval_start(tc, "Semaphore.tryAcquire");
+    r = uv_sem_trywait(sem->body.sem);
+    MVM_telemetry_interval_stop((MVMThreadContext*)(uintptr_t)r, interval_id, "tryAcquire result");
     return !r;
 }
 
 void MVM_semaphore_acquire(MVMThreadContext *tc, MVMSemaphore *sem) {
-    uv_sem_wait(&sem->body.sem);
+    unsigned int interval_id;
+    interval_id = MVM_telemetry_interval_start(tc, "Semaphore.acquire");
+    MVMROOT(tc, sem) {
+        MVM_gc_mark_thread_blocked(tc);
+        uv_sem_wait(sem->body.sem);
+        MVM_gc_mark_thread_unblocked(tc);
+    }
+    MVM_telemetry_interval_stop(tc, interval_id, "Semaphore.acquire");
 }
 
 void MVM_semaphore_release(MVMThreadContext *tc, MVMSemaphore *sem) {
-    uv_sem_post(&sem->body.sem);
+    MVM_telemetry_timestamp(tc, "Semaphore.release");
+    uv_sem_post(sem->body.sem);
 }
