@@ -119,6 +119,123 @@ MVMint16 MVM_nativecall_get_arg_type(MVMThreadContext *tc, MVMObject *info, MVMi
     return result;
 }
 
+MVMint16 determine_type_from_arg(MVMThreadContext *tc, MVMArgs args, MVMint16 pos, unsigned int interval_id) {
+    if (args.callsite->arg_flags[pos + 1] & MVM_CALLSITE_ARG_OBJ) {
+        MVMObject *value = args.source[args.map[pos + 1]].o;
+        if (!value) {
+            MVM_telemetry_interval_stop(tc, interval_id, "nativecall dispatch failed");
+            MVM_exception_throw_adhoc(tc, "Internal error: missing object for argument %d in MVM_nativecall_dispatch",
+                pos);
+        }
+        else if (!IS_CONCRETE(value)) {
+            MVM_telemetry_interval_stop(tc, interval_id, "nativecall dispatch failed");
+            MVM_exception_throw_adhoc(tc, "Type object for argument %d is not allowed", pos);
+        }
+        switch (REPR(value)->ID) {
+            case MVM_REPR_ID_P6int:
+                if (MVMP6int_is_unsigned(STABLE(value))) {
+                    return MVM_NATIVECALL_ARG_ULONG;
+                }
+                else {
+                    return MVM_NATIVECALL_ARG_LONG;
+                }
+            case MVM_REPR_ID_P6num:
+                return MVM_NATIVECALL_ARG_DOUBLE;
+            case MVM_REPR_ID_P6str:
+                return MVM_NATIVECALL_ARG_UTF8STR;
+
+            case MVM_REPR_ID_NativeRef:
+                ; // Need that semicolon to not trip up the compiler which
+                  // doesn't like declarations after switch labels.
+                MVMNativeRefREPRData *repr_data = (MVMNativeRefREPRData *)STABLE(value)->REPR_data;
+                switch (repr_data->primitive_type) {
+                    case MVM_STORAGE_SPEC_BP_INT8:
+                    case MVM_STORAGE_SPEC_BP_INT16:
+                    case MVM_STORAGE_SPEC_BP_INT32:
+                    case MVM_STORAGE_SPEC_BP_INT:
+                        return MVM_NATIVECALL_ARG_LONG;
+                    case MVM_STORAGE_SPEC_BP_UINT8:
+                    case MVM_STORAGE_SPEC_BP_UINT16:
+                    case MVM_STORAGE_SPEC_BP_UINT32:
+                    case MVM_STORAGE_SPEC_BP_UINT64:
+                        return MVM_NATIVECALL_ARG_ULONG;
+                    case MVM_STORAGE_SPEC_BP_NUM:
+                        return MVM_NATIVECALL_ARG_DOUBLE;
+                    case MVM_STORAGE_SPEC_BP_STR:
+                        return MVM_NATIVECALL_ARG_UTF8STR;
+                    default:
+                        MVM_telemetry_interval_stop(tc, interval_id, "nativecall dispatch failed");
+                        MVM_exception_throw_adhoc(tc, "Internal error: unexpected storage spec %d "
+                            " processing argument %d in MVM_nativecall_dispatch",
+                            repr_data->primitive_type, pos);
+                        break;
+                }
+            case MVM_REPR_ID_P6opaque:
+                MVM_telemetry_interval_stop(tc, interval_id, "nativecall dispatch failed");
+                MVM_exception_throw_adhoc(tc, "Incompatible high level Raku object seen when processing argument %d",
+                    pos);
+                break;
+
+            case MVM_REPR_ID_MVMCArray:
+                return MVM_NATIVECALL_ARG_CARRAY;
+
+            case MVM_REPR_ID_MVMCStruct:
+                return MVM_NATIVECALL_ARG_CSTRUCT;
+
+            case MVM_REPR_ID_MVMCUnion:
+                return MVM_NATIVECALL_ARG_CUNION;
+
+            case MVM_REPR_ID_MVMString:
+            case MVM_REPR_ID_VMArray:
+            case MVM_REPR_ID_MVMCode:
+            case MVM_REPR_ID_MVMCPointer:
+            case MVM_REPR_ID_MVMCStr:
+            case MVM_REPR_ID_MVMNull:
+            case MVM_REPR_ID_MVMCPPStruct:
+                MVM_telemetry_interval_stop(tc, interval_id, "nativecall dispatch failed");
+                MVM_exception_throw_adhoc(tc, "Internal error: NYI REPR type %d "
+                    " processing argument %d in MVM_nativecall_dispatch",
+                    REPR(value)->ID, pos);
+                break;
+            default:
+                MVM_telemetry_interval_stop(tc, interval_id, "nativecall dispatch failed");
+                MVM_exception_throw_adhoc(tc, "Internal error: unexpected REPR type %d "
+                    " processing argument %d in MVM_nativecall_dispatch",
+                    REPR(value)->ID, pos);
+                break;
+        }
+    }
+    else if (args.callsite->arg_flags[pos + 1] & MVM_CALLSITE_ARG_INT) {
+        return MVM_NATIVECALL_ARG_LONG;
+    }
+    else if (args.callsite->arg_flags[pos + 1] & MVM_CALLSITE_ARG_UINT) {
+        return MVM_NATIVECALL_ARG_ULONG;
+    }
+    else if (args.callsite->arg_flags[pos + 1] & MVM_CALLSITE_ARG_NUM) {
+        return MVM_NATIVECALL_ARG_DOUBLE;
+    }
+    else if (args.callsite->arg_flags[pos + 1] & MVM_CALLSITE_ARG_STR) {
+        return MVM_NATIVECALL_ARG_UTF8STR;
+    }
+    else {
+        MVM_telemetry_interval_stop(tc, interval_id, "nativecall invoke failed");
+        MVM_exception_throw_adhoc(tc, "Internal error: unhandled dyncall argument flags %d "
+            " processing argument %d in MVM_nativecall_dispatch",
+            args.callsite->arg_flags[pos + 1], pos);
+    }
+
+    // Should never be reached.
+    return 0;
+}
+
+void MVM_nativecall_fill_var_arg_types(MVMThreadContext *tc, MVMArgs args, MVMuint64 variadic_rw_bitfield, MVMint16 *arg_types, MVMint16 i, MVMint16 num_args, unsigned int interval_id) {
+    for (; i < num_args; i++) {
+        arg_types[i] = determine_type_from_arg(tc, args, i, interval_id);
+        if (variadic_rw_bitfield & 1<<i)
+            arg_types[i] |= MVM_NATIVECALL_ARG_RW_MASK;
+    }
+}
+
 MVMObject * MVM_nativecall_make_int(MVMThreadContext *tc, MVMObject *type, MVMint64 value) {
     return type ? MVM_repr_box_int(tc, type, value) : NULL;
 }
@@ -282,38 +399,49 @@ double MVM_nativecall_unmarshal_double(MVMThreadContext *tc, MVMObject *value) {
     return (double)MVM_repr_get_num(tc, value);
 }
 
+char * MVM_nativecall_encode_string(MVMThreadContext *tc, MVMString *value_str, MVMint16 type, MVMint16 *free, MVMint16 unmarshal_kind, const MVMREPROps *repr) {
+    /* Encode string. */
+    char *str;
+    switch (type & MVM_NATIVECALL_ARG_TYPE_MASK) {
+        case MVM_NATIVECALL_ARG_ASCIISTR:
+            str = MVM_string_ascii_encode_malloc(tc, value_str);
+            break;
+        case MVM_NATIVECALL_ARG_UTF16STR:
+        {
+            MVMuint64 output_size;
+            /* This adds a (two byte) U-0000 at the end, but doesn't include
+             * that code point in the returned output_size. */
+            char *temp = MVM_string_utf16_encode_substr(tc, value_str, &output_size, 0, -1, NULL, 0);
+            /* output_size is in 2 byte units, and doesn't include the
+             * terminating U-0000, which was written. */
+            size_t octet_len = 2 * (output_size + 1);
+            str = malloc(octet_len);
+            memcpy(str, temp, octet_len);
+            MVM_free(temp);
+            break;
+        }
+        default:
+            str = MVM_string_utf8_encode_C_string_malloc(tc, value_str);
+    }
+
+
+    /* Set whether to free it or not. */
+    if (free) {
+        if (repr->ID == MVM_REPR_ID_MVMCStr)
+            *free = 0; /* Manually managed. */
+        else if (type & MVM_NATIVECALL_ARG_FREE_STR_MASK)
+            *free = 1;
+        else
+            *free = 0;
+    }
+
+    return str;
+}
+
 char * MVM_nativecall_unmarshal_string(MVMThreadContext *tc, MVMObject *value, MVMint16 type, MVMint16 *free, MVMint16 unmarshal_kind) {
-    if (IS_CONCRETE(value)) {
-        MVMString *value_str = MVM_repr_get_str(tc, value);
-
-        /* Encode string. */
-        char *str;
-        switch (type & MVM_NATIVECALL_ARG_TYPE_MASK) {
-            case MVM_NATIVECALL_ARG_ASCIISTR:
-                str = MVM_string_ascii_encode_any(tc, value_str);
-                break;
-            case MVM_NATIVECALL_ARG_UTF16STR:
-                str = MVM_string_utf16_encode(tc, value_str, 0);
-                break;
-            default:
-                str = MVM_string_utf8_encode_C_string(tc, value_str);
-        }
-
-        /* Set whether to free it or not. */
-        if (free) {
-            if (REPR(value)->ID == MVM_REPR_ID_MVMCStr)
-                *free = 0; /* Manually managed. */
-            else if (free && type & MVM_NATIVECALL_ARG_FREE_STR_MASK)
-                *free = 1;
-            else
-                *free = 0;
-        }
-
-        return str;
-    }
-    else {
-        return NULL;
-    }
+    return IS_CONCRETE(value)
+        ? MVM_nativecall_encode_string(tc, MVM_repr_get_str(tc, value), type, free, unmarshal_kind, REPR(value))
+        : NULL;
 }
 
 MVM_NO_RETURN static void unmarshal_error(MVMThreadContext *tc, char *desired_repr, MVMObject *value, MVMint16 unmarshal_kind) MVM_NO_RETURN_ATTRIBUTE;
@@ -406,234 +534,6 @@ static const char *dlerror(void)
 }
 #endif
 
-void init_c_call_node(MVMThreadContext *tc, MVMSpeshGraph *sg, MVMJitNode *node, void *func_ptr, MVMint16 num_args, MVMJitCallArg *args) {
-    node->type = MVM_JIT_NODE_CALL_C;
-    node->u.call.func_ptr = func_ptr;
-    if (0 < num_args) {
-        node->u.call.args = MVM_spesh_alloc(tc, sg, num_args * sizeof(MVMJitCallArg));
-        memcpy(node->u.call.args, args, num_args * sizeof(MVMJitCallArg));
-    }
-    else {
-        node->u.call.args = NULL;
-    }
-    node->u.call.num_args = num_args;
-    node->u.call.rv_mode = MVM_JIT_RV_VOID;
-    node->u.call.rv_idx  = -1;
-}
-
-void save_rv_to_stack(MVMThreadContext *tc, MVMJitNode *node, MVMint32 storage_pos) {
-    node->u.call.rv_mode = MVM_JIT_RV_TO_STACK;
-    node->u.call.rv_idx = storage_pos;
-}
-
-void init_box_call_node(MVMThreadContext *tc, MVMSpeshGraph *sg, MVMJitNode *box_rv_node, void *func_ptr, MVMint16 restype, MVMint16 dst) {
-    MVMJitCallArg args[] = { { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } },
-                             { MVM_JIT_REG_DYNIDX, { 2 } },
-                             { MVM_JIT_STACK_VALUE, { 0 } }};
-    init_c_call_node(tc, sg, box_rv_node, func_ptr, 3, args);
-    box_rv_node->next = NULL;
-    if (dst == -1) {
-        box_rv_node->u.call.rv_mode = MVM_JIT_RV_DYNIDX;
-        box_rv_node->u.call.rv_idx = 0;
-    }
-    else {
-        box_rv_node->u.call.args[1].type = MVM_JIT_REG_VAL;
-        box_rv_node->u.call.args[1].v.reg = restype;
-
-        box_rv_node->u.call.rv_mode = MVM_JIT_RV_PTR;
-        box_rv_node->u.call.rv_idx = dst;
-    }
-}
-
-MVMJitGraph *MVM_nativecall_jit_graph_for_caller_code(
-    MVMThreadContext   *tc,
-    MVMSpeshGraph      *sg,
-    MVMNativeCallBody  *body,
-    MVMint16            restype,
-    MVMint16            dst,
-    MVMSpeshIns       **arg_ins
-) {
-    MVMJitGraph *jg = MVM_spesh_alloc(tc, sg, sizeof(MVMJitGraph)); /* will actually calloc */
-    MVMJitNode *block_gc_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-    MVMJitNode *unblock_gc_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-    MVMJitNode *call_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-    MVMJitNode *box_rv_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-    MVMJitCallArg block_gc_args[] = { { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } } };
-
-    jg->sg = sg;
-    jg->first_node = block_gc_node;
-    init_c_call_node(tc, sg, block_gc_node,   &MVM_gc_mark_thread_blocked, 1, block_gc_args);
-    block_gc_node->next = call_node;
-    init_c_call_node(tc, sg, call_node, body->entry_point, 0, NULL); /* we handle args manually */
-    save_rv_to_stack(tc, call_node, 0);
-
-    init_c_call_node(tc, sg, unblock_gc_node, &MVM_gc_mark_thread_unblocked, 1, block_gc_args);
-
-    call_node->next = unblock_gc_node;
-    call_node->u.call.num_args = body->num_args;
-    jg->last_node = unblock_gc_node->next = box_rv_node;
-
-    if (0 < body->num_args) {
-        MVMuint16 i = 0, str_arg_count = 0;
-        call_node->u.call.args = MVM_spesh_alloc(tc, sg, body->num_args * sizeof(MVMJitCallArg));
-        for (i = 0; i < body->num_args; i++) {
-            if ((body->arg_types[i] & MVM_NATIVECALL_ARG_TYPE_MASK) == MVM_NATIVECALL_ARG_UTF8STR) {
-                MVMJitNode *unbox_str_node;
-                MVMJitNode *free_str_node;
-
-                if (7 < ++str_arg_count) /* only got 7 empty slots in the stack scratch space */
-                    goto fail;
-
-                unbox_str_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-                {
-                    MVMJitCallArg unbox_str_args[] = {
-                        { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } },
-                        {
-                            dst == -1 ? MVM_JIT_ARG_I64 : MVM_JIT_PARAM_I64 ,
-                            { dst == -1 ? i : arg_ins[i]->operands[1].reg.orig }
-                        }
-                    };
-                    init_c_call_node(tc, sg, unbox_str_node, &MVM_string_utf8_maybe_encode_C_string, 2, unbox_str_args);
-                    save_rv_to_stack(tc, unbox_str_node, str_arg_count);
-                }
-                unbox_str_node->next = jg->first_node;
-                jg->first_node = unbox_str_node;
-
-                call_node->u.call.args[i].type = MVM_JIT_STACK_VALUE;
-                call_node->u.call.args[i].v.lit_i64 = str_arg_count;
-
-                if ((body->arg_types[i] & MVM_NATIVECALL_ARG_FREE_STR_MASK) != 0) {
-                    MVMJitCallArg free_str_args[] = {
-                        { MVM_JIT_STACK_VALUE , { str_arg_count } }
-                    };
-                    free_str_node = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-                    init_c_call_node(tc, sg, free_str_node, &MVM_free, 1, free_str_args);
-                    free_str_node->next = unblock_gc_node->next;
-                    unblock_gc_node->next = free_str_node;
-                }
-            }
-        }
-        for (i = 0; i < body->num_args; i++) {
-            MVMJitArgType arg_type;
-            int is_rw = ((body->arg_types[i] & MVM_NATIVECALL_ARG_RW_MASK) == MVM_NATIVECALL_ARG_RW);
-
-            switch (body->arg_types[i] & MVM_NATIVECALL_ARG_TYPE_MASK) {
-                case MVM_NATIVECALL_ARG_CHAR:
-                case MVM_NATIVECALL_ARG_UCHAR:
-                case MVM_NATIVECALL_ARG_SHORT:
-                case MVM_NATIVECALL_ARG_USHORT:
-                case MVM_NATIVECALL_ARG_INT:
-                case MVM_NATIVECALL_ARG_UINT:
-                case MVM_NATIVECALL_ARG_LONG:
-                case MVM_NATIVECALL_ARG_ULONG:
-                case MVM_NATIVECALL_ARG_LONGLONG:
-                case MVM_NATIVECALL_ARG_ULONGLONG:
-                    arg_type = dst == -1
-                        ? is_rw ? MVM_JIT_ARG_I64_RW : MVM_JIT_ARG_I64
-                        : is_rw ? MVM_JIT_PARAM_I64_RW : MVM_JIT_PARAM_I64;
-                    break;
-                case MVM_NATIVECALL_ARG_DOUBLE:
-                    if (is_rw) goto fail;
-                    arg_type = dst == -1
-                        ? MVM_JIT_ARG_DOUBLE
-                        : MVM_JIT_PARAM_DOUBLE;
-                    break;
-                case MVM_NATIVECALL_ARG_CPOINTER:
-                    if (is_rw) goto fail;
-                    arg_type = dst == -1 ? MVM_JIT_ARG_PTR : MVM_JIT_PARAM_PTR;
-                    break;
-                case MVM_NATIVECALL_ARG_CARRAY:
-                    if (body->arg_types[i] & MVM_NATIVECALL_ARG_REFRESH_MASK) goto fail;
-                    if (is_rw) goto fail;
-                    arg_type = dst == -1 ? MVM_JIT_ARG_PTR : MVM_JIT_PARAM_PTR;
-                    break;
-                case MVM_NATIVECALL_ARG_VMARRAY:
-                    if (is_rw) goto fail;
-                    arg_type = dst == -1 ? MVM_JIT_ARG_VMARRAY : MVM_JIT_PARAM_VMARRAY;
-                    break;
-                case MVM_NATIVECALL_ARG_UTF8STR:
-                    if (is_rw) goto fail;
-                    continue; /* already handled */
-                default:
-                    goto fail;
-            }
-            call_node->u.call.args[i].type = arg_type;
-            call_node->u.call.args[i].v.lit_i64 = dst == -1 ? i : arg_ins[i]->operands[1].reg.orig;
-        }
-    }
-
-    if (body->ret_type == MVM_NATIVECALL_ARG_CHAR
-        || body->ret_type == MVM_NATIVECALL_ARG_UCHAR
-        || body->ret_type == MVM_NATIVECALL_ARG_SHORT
-        || body->ret_type == MVM_NATIVECALL_ARG_USHORT
-        || body->ret_type == MVM_NATIVECALL_ARG_INT
-        || body->ret_type == MVM_NATIVECALL_ARG_UINT
-        || body->ret_type == MVM_NATIVECALL_ARG_LONG
-        || body->ret_type == MVM_NATIVECALL_ARG_ULONG
-        || body->ret_type == MVM_NATIVECALL_ARG_LONGLONG
-        || body->ret_type == MVM_NATIVECALL_ARG_ULONGLONG
-    ) {
-        call_node->u.call.rv_type = body->ret_type;
-        init_box_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_int, restype, dst);
-    }
-    else if (body->ret_type == MVM_NATIVECALL_ARG_CPOINTER) {
-        init_box_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_cpointer, restype, dst);
-    }
-    else if (body->ret_type == MVM_NATIVECALL_ARG_UTF8STR) {
-        MVMJitCallArg args[] = { { MVM_JIT_INTERP_VAR , { MVM_JIT_INTERP_TC } },
-                                 { MVM_JIT_REG_DYNIDX, { 2 } },
-                                 { MVM_JIT_LITERAL, { MVM_NATIVECALL_ARG_UTF8STR } },
-                                 { MVM_JIT_STACK_VALUE, { 0 } }};
-        init_c_call_node(tc, sg, box_rv_node, &MVM_nativecall_make_str, 4, args);
-        box_rv_node->next = NULL;
-        if (dst == -1) {
-            box_rv_node->u.call.rv_mode = MVM_JIT_RV_DYNIDX;
-            box_rv_node->u.call.rv_idx = 0;
-        }
-        else {
-            box_rv_node->u.call.args[1].type = MVM_JIT_REG_VAL;
-            box_rv_node->u.call.args[1].v.reg = restype;
-
-            box_rv_node->u.call.rv_mode = MVM_JIT_RV_PTR;
-            box_rv_node->u.call.rv_idx = dst;
-        }
-    }
-    else if (body->ret_type == MVM_NATIVECALL_ARG_VOID) {
-        call_node->next = unblock_gc_node;
-        unblock_gc_node->next = NULL;
-        jg->last_node = unblock_gc_node;
-    }
-    else {
-        goto fail;
-    }
-
-    return jg;
-fail:
-    return NULL;
-}
-
-MVMJitCode *create_caller_code(MVMThreadContext *tc, MVMNativeCallBody *body) {
-    MVMSpeshGraph *sg = MVM_calloc(1, sizeof(MVMSpeshGraph));
-    MVMJitGraph *jg = MVM_nativecall_jit_graph_for_caller_code(tc, sg, body, -1, -1, NULL);
-    MVMJitCode *jitcode;
-    if (jg != NULL) {
-        MVMJitNode *entry_label = MVM_spesh_alloc(tc, sg, sizeof(MVMJitNode));
-        entry_label->next = jg->first_node;
-        jg->first_node = entry_label;
-        jg->num_labels = 1;
-        jg->no_trampoline = 1;
-
-        entry_label->type = MVM_JIT_NODE_LABEL;
-        entry_label->u.label.name = 0;
-        jitcode = MVM_jit_compile_graph(tc, jg);
-    }
-    else {
-        jitcode = NULL;
-    }
-    MVM_spesh_graph_destroy(tc, sg);
-    return jitcode;
-}
-
 void MVM_nativecall_setup(MVMThreadContext *tc, MVMNativeCallBody *body, unsigned int interval_id) {
     /* Try to load the library. */
     DLLib *lib_handle = MVM_nativecall_load_lib(body->lib_name[0] ? body->lib_name : NULL);
@@ -660,12 +560,6 @@ void MVM_nativecall_setup(MVMThreadContext *tc, MVMNativeCallBody *body, unsigne
         }
     }
 
-    if (tc->instance->jit_enabled) {
-        body->jitcode = create_caller_code(tc, body);
-    }
-    else
-        body->jitcode = NULL;
-
     /* Initialize body->lib_handle as late as possible, so get_int won't report
        that the NativeCall site is set up too early */
     body->lib_handle = lib_handle;
@@ -673,7 +567,7 @@ void MVM_nativecall_setup(MVMThreadContext *tc, MVMNativeCallBody *body, unsigne
 
 /* Builds up a native call site out of the supplied arguments. */
 MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *lib,
-        MVMString *sym, MVMString *conv, MVMObject *arg_info, MVMObject *ret_info) {
+        MVMString *sym, MVMString *conv, MVMObject *in_arg_info, MVMObject *ret_info) {
     char *lib_name = MVM_string_utf8_c8_encode_C_string(tc, lib);
     char *sym_name = MVM_string_utf8_c8_encode_C_string(tc, sym);
     MVMint8  keep_sym_name = 0;
@@ -687,13 +581,19 @@ MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *l
         tc->instance->str_consts.resolve_lib_name);
     MVMObject *resolve_lib_name_arg = (MVMObject *)MVM_repr_at_key_o(tc, ret_info,
         tc->instance->str_consts.resolve_lib_name_arg);
+    MVMint64 variadic = MVM_repr_exists_key(tc, ret_info,
+        tc->instance->str_consts.variadic);
 
     /* Initialize the object; grab native call part of its body. */
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
 
     body->lib_name = lib_name;
-    MVM_ASSIGN_REF(tc, &(site->header), body->resolve_lib_name, resolve_lib_name);
-    MVM_ASSIGN_REF(tc, &(site->header), body->resolve_lib_name_arg, resolve_lib_name_arg);
+    if (!MVM_is_null(tc, resolve_lib_name)) {
+        if (!MVM_code_iscode(tc, resolve_lib_name))
+            MVM_exception_throw_adhoc(tc, "resolve_lib_name must be a code handle");
+        MVM_ASSIGN_REF(tc, &(site->header), body->resolve_lib_name, resolve_lib_name);
+        MVM_ASSIGN_REF(tc, &(site->header), body->resolve_lib_name_arg, resolve_lib_name_arg);
+    }
 
     /* Try to locate the symbol. */
     if (entry_point_o && !MVM_is_null(tc, entry_point_o)) {
@@ -717,26 +617,31 @@ MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *l
     body->convention = MVM_nativecall_get_calling_convention(tc, conv);
 
     /* Transform each of the args info structures into a flag. */
-    body->num_args  = MVM_repr_elems(tc, arg_info);
-    body->arg_types = MVM_malloc(sizeof(MVMint16) * (body->num_args ? body->num_args : 1));
-    body->arg_info  = MVM_malloc(sizeof(MVMObject *) * (body->num_args ? body->num_args : 1));
+    MVMint16 num_args  = MVM_repr_elems(tc, in_arg_info);
+    MVMint16   *arg_types = MVM_malloc(sizeof(MVMint16) * (num_args ? num_args : 1));
+    MVMObject **arg_info  = MVM_malloc(sizeof(MVMObject *) * (num_args ? num_args : 1));
 #ifdef HAVE_LIBFFI
-    body->ffi_arg_types = MVM_malloc(sizeof(ffi_type *) * (body->num_args ? body->num_args : 1));
+    body->ffi_arg_types = MVM_malloc(sizeof(ffi_type *) * (num_args ? num_args : 1));
 #endif
-    for (i = 0; i < body->num_args; i++) {
-        MVMObject *info = MVM_repr_at_pos_o(tc, arg_info, i);
-        body->arg_types[i] = MVM_nativecall_get_arg_type(tc, info, 0);
+    for (i = 0; i < num_args; i++) {
+        MVMObject *info = MVM_repr_at_pos_o(tc, in_arg_info, i);
+        arg_types[i] = MVM_nativecall_get_arg_type(tc, info, 0);
 #ifdef HAVE_LIBFFI
-        body->ffi_arg_types[i] = MVM_nativecall_get_ffi_type(tc, body->arg_types[i]);
+        body->ffi_arg_types[i] = MVM_nativecall_get_ffi_type(tc, arg_types[i]);
 #endif
-        if(body->arg_types[i] == MVM_NATIVECALL_ARG_CALLBACK) {
-            MVM_ASSIGN_REF(tc, &(site->header), body->arg_info[i],
+        if(arg_types[i] == MVM_NATIVECALL_ARG_CALLBACK) {
+            MVM_ASSIGN_REF(tc, &(site->header), arg_info[i],
                 MVM_repr_at_key_o(tc, info, tc->instance->str_consts.callback_args));
         }
         else {
-            body->arg_info[i]  = NULL;
+            arg_info[i]  = NULL;
         }
     }
+    body->arg_types = arg_types;
+    body->arg_info  = arg_info;
+    MVM_barrier();
+    body->num_args  = num_args; /* ensure we have properly populated arrays before setting num */
+    body->variadic = variadic;
 
     /* Transform return argument type info a flag. */
     body->ret_type     = MVM_nativecall_get_arg_type(tc, ret_info, 1);
@@ -748,13 +653,13 @@ MVMint8 MVM_nativecall_build(MVMThreadContext *tc, MVMObject *site, MVMString *l
 
     MVM_telemetry_interval_stop(tc, interval_id, "nativecall built");
 
-    return body->jitcode != NULL;
+    return 0;
 }
 
 static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec, MVMObject *target_type, void *cpointer_body) {
     MVMObject *result = NULL;
 
-    MVMROOT2(tc, target_spec, target_type, {
+    MVMROOT2(tc, target_spec, target_type) {
         switch (REPR(target_type)->ID) {
             case MVM_REPR_ID_P6opaque: {
                 const MVMStorageSpec *ss = REPR(target_spec)->get_storage_spec(tc, STABLE(target_spec));
@@ -892,7 +797,7 @@ static MVMObject * nativecall_cast(MVMThreadContext *tc, MVMObject *target_spec,
             default:
                 MVM_exception_throw_adhoc(tc, "Internal error: unhandled target type");
         }
-    });
+    }
 
     return result;
 }
@@ -1150,24 +1055,23 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
 
 typedef struct ResolverData {
     MVMObject *site;
-    MVMRegister args[1];
 } ResolverData;
 static void callback_invoke(MVMThreadContext *tc, void *data) {
     /* Invoke the coderef, to set up the nested interpreter. */
     ResolverData *r = (ResolverData*)data;
     MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, r->site);
-    MVMObject *code = body->resolve_lib_name;
-    MVMCallsite *callsite = MVM_callsite_get_common(tc, MVM_CALLSITE_ID_INV_ARG);
-    r->args[0].o = body->resolve_lib_name_arg;
-    STABLE(code)->invoke(tc, code, callsite, r->args);
+    MVMCallStackArgsFromC *args_record = MVM_callstack_allocate_args_from_c(tc,
+            MVM_callsite_get_common(tc, MVM_CALLSITE_ID_OBJ));
+    args_record->args.source[0].o = body->resolve_lib_name_arg;
+    MVM_frame_dispatch(tc, body->resolve_lib_name, args_record->args, -1);
 
     /* Ensure we exit interp after callback. */
     tc->thread_entry_frame = tc->cur_frame;
 }
 void MVM_nativecall_restore_library(MVMThreadContext *tc, MVMNativeCallBody *body, MVMObject *root) {
-    if (!MVM_is_null(tc, body->resolve_lib_name) && !MVM_is_null(tc, body->resolve_lib_name_arg)) {
+    if (body->resolve_lib_name && !MVM_is_null(tc, body->resolve_lib_name_arg)) {
         MVMRegister res = {NULL};
-        ResolverData data = {root, {NULL}};
+        ResolverData data = {root};
 
         MVM_interp_run_nested(tc, callback_invoke, &data, &res);
 
@@ -1184,19 +1088,3 @@ void MVM_nativecall_restore_library(MVMThreadContext *tc, MVMNativeCallBody *bod
         MVM_nativecall_setup(tc, body, 0);
     }
 }
-
-void MVM_nativecall_invoke_jit(MVMThreadContext *tc, MVMObject *site) {
-    MVMNativeCallBody *body = MVM_nativecall_get_nc_body(tc, site);
-
-    if (MVM_UNLIKELY(!body->lib_handle)) {
-        MVMROOT(tc, site, {
-            MVM_nativecall_restore_library(tc, body, site);
-        });
-        body = MVM_nativecall_get_nc_body(tc, site);
-    }
-
-    MVMJitCode * const jitcode = body->jitcode;
-    assert(jitcode);
-    assert(jitcode->func_ptr);
-    jitcode->func_ptr(tc, *tc->interp_cu, jitcode->labels[0]);
-}
